import random

from simple_playgrounds.playground import PlaygroundRegister
from simple_playgrounds.playgrounds.empty import ConnectedRooms2D, SingleRoom
from simple_playgrounds.playgrounds.scene_elements import (Basic,
                                                           DeathZone,
                                                           GoalZone,
                                                           )
from simple_playgrounds.utils.position_utils import CoordinateSampler


@PlaygroundRegister.register('navigation', 'endgoal_cue')
class EndgoalRoomCue(SingleRoom):
    """
    Squared environment with obstacles.
    The agent must reach the correct corner, indicated by the color of
    a center object.
    """

    def __init__(self,
                 time_limit=1000,
                 reward_reached_time_limit=-10,
                 reward_reached_endgoal=10,
                 reward_reached_deathtrap=-10,
                 wall_texture_seed = None,
                 ):

        super().__init__(size=(200, 200), wall_texture_seed=wall_texture_seed)

        # Starting area of the agent
        area_center, _ = self.area_rooms[(0, 0)]
        area_start = CoordinateSampler(center=area_center,
                                       area_shape='rectangle',
                                       width_length=(100, 100))
        self.initial_agent_coordinates = area_start

        # Visual Cues for the agent to orient itself.
        obstacle_1 = Basic(default_config_key='pentagon', radius=9)
        self.add_scene_element(obstacle_1, ((60, 30), 0.34))

        obstacle_2 = Basic(default_config_key='rectangle', width_length=[8, 12])
        self.add_scene_element(obstacle_2, ((130, 150), 1.7))

        obstacle_3 = Basic(default_config_key='square', radius=8)
        self.add_scene_element(obstacle_3, ((40, 140), 0.4))

        obstacle_4 = Basic(physical_shape='triangle', radius=14, texture=(150, 200, 200))
        self.add_scene_element(obstacle_4, ((160, 60), 0))

        self.goal = None
        self.cue = None

        self.goal_locations = (((20, 20), 0), ((180, 20), 0), ((180, 180), 0), ((20, 180), 0))
        self.cue_colors = ((200, 50, 200), (50, 200, 50), (200, 50, 50), (50, 50, 200))

        self.reward_goal = reward_reached_endgoal
        self.reward_deathtrap = reward_reached_deathtrap

        self._set_goal()

        self.time_limit = time_limit
        self.time_limit_reached_reward = reward_reached_time_limit

        self.create_entity_types_map([Basic])

    def _set_goal(self):

        index_goal = random.randint(0, 3)
        loc = self.goal_locations[index_goal]
        col = self.cue_colors[index_goal]

        if self.goal is not None:
            self.remove_scene_element(self.goal)
            self.remove_scene_element(self.cue)

        self.cue = Basic(physical_shape='circle', radius=10, texture=col, is_temporary_entity=True)
        self.add_scene_element(self.cue, ((100, 100), 0))

        self.goal = GoalZone(reward=self.reward_goal, is_temporary_entity=True)
        self.add_scene_element(self.goal, loc)

        for i in range(4):
            if i != index_goal:
                loc = self.goal_locations[i]
                other_goal = DeathZone(reward=self.reward_deathtrap, is_temporary_entity=True)
                self.add_scene_element(other_goal, loc)

    def reset(self):
        super().reset()
        self._set_goal()


@PlaygroundRegister.register('navigation', 'endgoal_9rooms')
class Endgoal9Rooms(ConnectedRooms2D):
    """
    Squared environment composed of 9 rooms (3x3).
    The agent must reach the invisible goal in the left-down corner.
    Each wall has a different color.
    """
    def __init__(self,
                 time_limit = 1000,
                 wall_texture_seed = None):

        super().__init__(size=(450, 450),
                         room_layout=(3, 3),
                         wall_type='colorful',
                         wall_texture_seed=wall_texture_seed)

        # Starting area of the agent
        area_start = CoordinateSampler(center=(225, 225),
                                       area_shape='rectangle', width_length=(450, 450))
        self.initial_agent_coordinates = area_start

        # invisible endzone at one corner of the game
        invisible_endzone = GoalZone(reward=10)
        self.add_scene_element(invisible_endzone, ((20, 20), 0))

        self.time_limit = time_limit

        self.create_entity_types_map()
